/* $Id: api.c 32 2023-04-26 16:49:22Z nishi $ */
#include <swinehub/api.h>

#include <swinehub.h>
#include <swinehub/config.h>
#include <swinehub/util.h>
#include <swinehub/sha512.h>
#include <swinehub/db.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>

#include <openssl/sha.h>

#define WRITE(fd, data) write(fd, data, strlen(data))
#define _FWRITE(f, data) fwrite(data, strlen(data), 1, f)

EXTERN_DB;

extern char* swinehub_body;

void swinehub_api(struct swinehub_config* config){
	char* path = getenv("PATH_REPLACED");
	if(strcmp(path, "/v1/repo/latest-log") == 0){
		printf("Content-Type: text/plain\r\n");
	}else{
		printf("Content-Type: application/json\r\n");
	}
	if(strcmp(path, "/v1") == 0){
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		printf("{\"message\": \"API is working\", \"code\": %d}", SWINEHUB_CODE_OK);
	}else if(strcmp(path, "/v1/version") == 0){
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		printf("{\"version\": \"%s\", \"code\": %d}", SWINEHUB_VERSION, SWINEHUB_CODE_OK);
	}else if(strcmp(path, "/v1/valid") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Valid token\", \"username\": \"%s\", \"code\": %d}", result.dptr, SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/repos") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr != NULL){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			char* repos = malloc(1);
			repos[0] = 0;
			DIR* dir = opendir(config->path);
			struct dirent* d;
			char* prefix = sstrcat(swinehub_body, "-");
			while((d = readdir(dir)) != NULL){
				int i;
				char cond = 1;
				for(i = 0; d->d_name[i] != 0 && prefix[i] != 0; i++){
					if(prefix[i] != d->d_name[i]){
						cond = 0;
						break;
					}
				}
				if(cond){
					char* tmp = repos;
					repos = sstrcat5(tmp, repos[0] == 0 ? "" : ", ", "\"", d->d_name + strlen(prefix), "\"");
					free(tmp);
				}
			}
			free(prefix);
			printf("{\"repos\": [%s], \"code\": %d}", repos, SWINEHUB_CODE_OK);
		}else{
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"User does not exist\", \"code\": %d}", SWINEHUB_CODE_USER_DOES_NOT_EXIST);
		}
	}else if(strcmp(path, "/v1/logout") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		int err = dbm_delete(swinehub_token_db, key);
		swinehub_db_init();
		if(err == 0){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Logout successful\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/signup") == 0){
		if(!config->signup){
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Signup disabled\", \"code\": %d}", SWINEHUB_CODE_SIGNUP_DISABLED);
			return;
		}
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* password = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:password`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		if(strlen(username) > SWINEHUB_USERNAME_MAX){
			free(backup);
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Username too long (>%d)\", \"code\": %d}", SWINEHUB_USERNAME_MAX, SWINEHUB_CODE_USERNAME_TOO_LONG);
			return;
		}
		int hi;
		for(hi = 0; username[hi] != 0; hi++){
			if(username[hi] == '-'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept hyphens\", \"code\": %d}", SWINEHUB_CODE_NO_HYPHENS);
				return;
			}else if(username[hi] == '#'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept hashes\", \"code\": %d}", SWINEHUB_CODE_NO_HASHES);
				return;
			}else if(username[hi] == '['){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept brackets\", \"code\": %d}", SWINEHUB_CODE_NO_BRACKETS);
				return;
			}else if(username[hi] == ']'){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept brackets\", \"code\": %d}", SWINEHUB_CODE_NO_BRACKETS);
				return;
			}else if(username[hi] == '='){
				printf("Status: 400 Bad Request\r\n");
				printf("\r\n");
				printf("{\"error\": \"Username does not accept equals\", \"code\": %d}", SWINEHUB_CODE_NO_EQUALS);
				return;
			}
		}
		free(backup);
		datum key;
		key.dptr = username;
		key.dsize = strlen(username) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr != NULL){
			printf("Status: 409 Conflict\r\n");
			printf("\r\n");
			struct swinehub_user* usr = (struct swinehub_user*)result.dptr;
			printf("{\"error\": \"User exists\", \"code\": %d}", SWINEHUB_CODE_USER_EXISTS);
		}else{
			key.dptr = username;
			key.dsize = strlen(username) + 1;
			datum value;
			struct swinehub_user user;
			memset(user.username, 0, SWINEHUB_USERNAME_MAX + 1);
			strcpy(user.username, username);
			unsigned char* sha512passwd = swinehub_sha512(password);
			memcpy(user.password, sha512passwd, SHA512_DIGEST_LENGTH);
			free(sha512passwd);
			value.dptr = (void*)&user;
			value.dsize = sizeof(user);
			if(0 != dbm_store(swinehub_user_db, key, value, DBM_REPLACE)){
				printf("Status: 500 Internal Server Error\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
			}else{
				int pid = fork();
				if(pid == 0){
					execl(config->htpasswd, config->htpasswd, "-b", SWINEHUB_AUTH, user.username, password, NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
				printf("Status: 200 OK\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database successfully updated\", \"code\": %d}", SWINEHUB_CODE_OK);
			}
			swinehub_db_init();
		}
	}else if(strcmp(path, "/v1/login") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* password = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:password`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		datum key;
		key.dptr = username;
		key.dsize = strlen(username) + 1;
		datum result = dbm_fetch(swinehub_user_db, key);
		if(result.dptr == NULL){
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"User does not exist\", \"code\": %d}", SWINEHUB_CODE_USER_DOES_NOT_EXIST);
		}else{
			struct swinehub_user* usr = (struct swinehub_user*)result.dptr;
			char* passwd = swinehub_sha512(password);
			if(memcmp(usr->password, passwd, SHA512_DIGEST_LENGTH) == 0){
				char* token;
				char* buf;
				FILE* f;
				int i;
				while(1){
					token = malloc(1);
					token[0] = 0;
					buf = malloc(3);
					buf[2] = 0;
					f = fopen("/dev/urandom", "r");
					for(i = 0; i < 64; i++){
						unsigned char value;
						fread(&value, 2, 1, f);
						sprintf(buf, "%.2x", value);
						char* tmp = token;
						token = sstrcat(tmp, buf);
						free(tmp);
					}
					free(buf);
					fclose(f);
					key.dptr = token;
					key.dsize = strlen(token) + 1;
					result = dbm_fetch(swinehub_token_db, key);
					if(result.dptr != NULL){
						free(token);
					}else{
						break;
					}
				}
				key.dptr = token;
				key.dsize = strlen(token) + 1;
				datum value;
				value.dptr = username;
				value.dsize = strlen(username) + 1;
				if(0 != dbm_store(swinehub_token_db, key, value, DBM_REPLACE)){
					printf("Status: 500 Internal Server Error\r\n");
					printf("\r\n");
					printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
				}else{
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					printf("{\"token\": \"%s\", \"code\": %d}", token, SWINEHUB_CODE_OK);
				}
				swinehub_db_init();
				free(token);
			}else{
				printf("Status: 403 Forbidden\r\n");
				printf("\r\n");
				printf("{\"error\": \"Incorrect Password\", \"code\": %d}", SWINEHUB_CODE_INCORRECT_PASSWORD);
			}
			free(passwd);
		}
	}else if(strcmp(path, "/v1/repo") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* tmp = sstrcat3(config->path, "/", username);
		char* str = sstrcat3(tmp, "-", reponame);
		free(tmp);
		struct stat s;
		if(stat(str, &s) == 0){
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Found the repository\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
		}
		free(str);
	}else if(strcmp(path, "/v1/repo/revision") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
		char* reponame = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* tmp = sstrcat3(config->path, "/", username);
		char* str = sstrcat3(tmp, "-", reponame);
		free(tmp);
		char* rev = sstrcat(str, "/db/current");
		FILE* f = fopen(rev, "r");
		if(f != NULL){
			struct stat s;
			stat(rev, &s);
			char* revision = malloc(s.st_size);
			revision[s.st_size - 1] = 0;
			fread(revision, s.st_size - 1, 1, f);
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"revision\": %s, \"code\": %d}", revision, SWINEHUB_CODE_OK);
			fclose(f);
		}else{
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
		}
		free(str);
		free(rev);
	}else if(strcmp(path, "/v1/repo/latest-log") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* chdirpath = sstrcat5(config->cachepath, "/", username, "-", reponame);
		char* repopath = sstrcat5(config->path, "/", username, "-", reponame);
		struct stat st;
		if(stat(repopath, &st) != 0){
			free(repopath);
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
			return;
		}
		free(repopath);
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		fflush(stdout);
		if(stat(chdirpath, &st) != 0){
			int pid = fork();
			if(pid == 0){
				int nullfd = open("/dev/null", O_RDWR);
				dup2(nullfd, fileno(stdout));
				dup2(nullfd, fileno(stderr));
				chdir(config->cachepath);
				char* _filepath = sstrcat5("file:///", config->path, "/", username, "-");
				char* filepath = sstrcat(_filepath, reponame);
				free(_filepath);
				execl(config->svn, config->svn, "co", filepath, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
		}
		int pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stdout));
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "up", NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "log", "-r", "HEAD", NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		free(chdirpath);
	}else if(strcmp(path, "/v1/repo/set-readme") == 0){
		datum key;
		key.dptr = getenv("HTTP_X_TOKEN");
		key.dsize = strlen(getenv("HTTP_X_TOKEN")) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* repopath = sstrcat6(config->path, "/", (char*)result.dptr, "-", getenv("HTTP_X_REPONAME"), "/README.txt");
			FILE* f = fopen(repopath, "w");
			if(f != NULL){
				_FWRITE(f, swinehub_body);
				fclose(f);
				printf("Status: 200 OK\r\n");
				printf("\r\n");
				printf("{\"message\": \"README updated successfully\", \"code\": %d}", SWINEHUB_CODE_OK);
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
			}
			free(repopath);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/user/delete") == 0){
		datum key;
		key.dptr = swinehub_body;
		key.dsize = strlen(swinehub_body) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* username = strdup(result.dptr);
			key.dptr = result.dptr;
			key.dsize = strlen(result.dptr) + 1;
			if(0 != dbm_delete(swinehub_user_db, key)){
				printf("Status: 500 Internal Server Error\r\n");
				printf("\r\n");
				printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
			}else{
				swinehub_db_init();
				key.dptr = swinehub_body;
				key.dsize = strlen(swinehub_body) + 1;
				if(0 != dbm_delete(swinehub_token_db, key)){
					printf("Status: 500 Internal Server Error\r\n");
					printf("\r\n");
					printf("{\"message\": \"Database update failed: %s\", \"code\": %d}", strerror(errno), SWINEHUB_CODE_DB);
				}else{
					swinehub_db_init();
					DIR* dir = opendir(config->path);
					if(dir != NULL){
						struct dirent* d;
						char* cmp = sstrcat(username, "-");
						while((d = readdir(dir)) != NULL){
							if(strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0) continue;
							if(strlen(d->d_name) < strlen(cmp)) continue;
							char* tmp = strdup(d->d_name);
							tmp[strlen(cmp)] = 0;
							if(strcmp(tmp, cmp) == 0){
								int pid = fork();
								if(pid == 0){
									chdir(config->path);
									execl("/bin/rm", "/bin/rm", "-rf", d->d_name, NULL);
									_exit(-1);
								}else{
									waitpid(pid, 0, 0);
								}
							}
							free(tmp);
						}
						closedir(dir);
					}
					dir = opendir(config->cachepath);
					if(dir != NULL){
						struct dirent* d;
						char* cmp = sstrcat(username, "-");
						while((d = readdir(dir)) != NULL){
							if(strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0) continue;
							if(strlen(d->d_name) < strlen(cmp)) continue;
							char* tmp = strdup(d->d_name);
							tmp[strlen(cmp)] = 0;
							if(strcmp(tmp, cmp) == 0){
								int pid = fork();
								if(pid == 0){
									chdir(config->cachepath);
									execl("/bin/rm", "/bin/rm", "-rf", d->d_name, NULL);
									_exit(-1);
								}else{
									waitpid(pid, 0, 0);
								}

							}
							free(tmp);
						}
						closedir(dir);
					}
					printf("Status: 200 OK\r\n");
					printf("\r\n");
					printf("{\"error\": \"User deleted successfully\", \"code\": %d}", SWINEHUB_CODE_OK);
				}
			}
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/repo/log") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* username = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		char* logpath;
		if(strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame:path`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		for(i = 0; reponame[i] != 0; i++){
			if(reponame[i] == ':'){
				reponame[i] = 0;
				break;
			}
		}
		logpath = reponame + i + 1;
		if(strlen(reponame) + 1 + strlen(username) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `username:reponame:path`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		char* chdirpath = sstrcat5(config->cachepath, "/", username, "-", reponame);
		char* repopath = sstrcat5(config->path, "/", username, "-", reponame);
		struct stat st;
		if(stat(repopath, &st) != 0){
			free(repopath);
			printf("Status: 404 Not Found\r\n");
			printf("\r\n");
			printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
			return;
		}
		free(repopath);
		printf("Status: 200 OK\r\n");
		printf("\r\n");
		fflush(stdout);
		if(stat(chdirpath, &st) != 0){
			int pid = fork();
			if(pid == 0){
				int nullfd = open("/dev/null", O_RDWR);
				dup2(nullfd, fileno(stdout));
				dup2(nullfd, fileno(stderr));
				chdir(config->cachepath);
				char* _filepath = sstrcat5("file:///", config->path, "/", username, "-");
				char* filepath = sstrcat(_filepath, reponame);
				free(_filepath);
				execl(config->svn, config->svn, "co", filepath, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
		}
		int pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stdout));
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "up", NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		char* filepath = sstrcat("./", logpath);
		pid = fork();
		if(pid == 0){
			int nullfd = open("/dev/null", O_RDWR);
			dup2(nullfd, fileno(stderr));
			chdir(chdirpath);
			execl(config->svn, config->svn, "log", "-r", "COMMITTED", filepath, NULL);
			_exit(-1);
		}else{
			waitpid(pid, 0, 0);
		}
		free(filepath);
		free(chdirpath);
	}else if(strcmp(path, "/v1/repo/create") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* token = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(token) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `token:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		datum key;
		key.dptr = token;
		key.dsize = strlen(token) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* actualpath = sstrcat5(config->path, "/", result.dptr, "-", reponame);
			struct stat st;
			if(stat(actualpath, &st) == 0){
				printf("Status: 403 Forbidden\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository already exists\", \"code\": %d}", SWINEHUB_CODE_REPO_ALREADY_EXISTS);
				free(actualpath);
				return;
			}
			free(actualpath);
			while(access(SWINEHUB_AUTHZ_LOCK, F_OK) == 0) usleep(100);
			fclose(fopen(SWINEHUB_AUTHZ_LOCK, "w"));
			FILE* authz = fopen(SWINEHUB_AUTHZ, "a");
			char* name = sstrcat4(result.dptr, "-", reponame, ":/");
			char* data = sstrcat5("\n[", name, "]\n* = r\n", result.dptr, " = rw\n");
			char* tmp = data;
			data = sstrcat6("\n# START ", result.dptr, "-", reponame, data, "# END\n");
			free(tmp);
			free(name);
			_FWRITE(authz, data);
			fclose(authz);
			free(data);
			remove(SWINEHUB_AUTHZ_LOCK);
			data = sstrcat3(result.dptr, "-", reponame);
			int pid = fork();
			if(pid == 0){
				chdir(config->path);
				execl(config->svnadmin, config->svnadmin, "create", data, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Repository successfully created\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}
	}else if(strcmp(path, "/v1/repo/delete") == 0){
		int i;
		char* backup = strdup(swinehub_body);
		for(i = 0; swinehub_body[i] != 0; i++){
			if(swinehub_body[i] == ':'){
				swinehub_body[i] = 0;
				break;
			}
		}
		char* token = swinehub_body;
		char* reponame = swinehub_body + i + 1;
		if(strlen(token) == strlen(backup)){
			free(backup);
			printf("Status: 400 Bad Request\r\n");
			printf("\r\n");
			printf("{\"error\": \"Required parameters missing, `token:reponame`\", \"code\": %d}", SWINEHUB_CODE_PARAM_MISSING);
			return;
		}
		free(backup);
		datum key;
		key.dptr = token;
		key.dsize = strlen(token) + 1;
		datum result = dbm_fetch(swinehub_token_db, key);
		if(result.dptr != NULL){
			char* actualpath = sstrcat5(config->path, "/", result.dptr, "-", reponame);
			struct stat st;
			if(stat(actualpath, &st) != 0){
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
				printf("{\"error\": \"Repository not found\", \"code\": %d}", SWINEHUB_CODE_REPO_NOT_FOUND);
				free(actualpath);
				return;
			}
			free(actualpath);
			while(access(SWINEHUB_AUTHZ_LOCK, F_OK) == 0) usleep(100);
			fclose(fopen(SWINEHUB_AUTHZ_LOCK, "w"));

			FILE* authz = fopen(SWINEHUB_AUTHZ, "r");
			char* authzbuf = malloc(1);
			authzbuf[0] = 0;
			char* buf = malloc(1);
			buf[0] = 0;
			char ch;
			int length = 0;
			char cut_str = 0;
			char* start_tag = sstrcat4("# START ", result.dptr, "-", reponame);
			while(1){
				fread(&ch, 1, 1, authz);
				if(feof(authz) || ch == '\n'){
					if(strcmp(buf, start_tag) == 0){
						cut_str = 1;
					}else if(strcmp(buf, "# END") == 0){
						cut_str = 0;
					}else if(cut_str != 1){
						char* tmp = authzbuf;
						authzbuf = sstrcat3(tmp, buf, "\n");
						free(tmp);
					}
					free(buf);
					buf = malloc(1);
					buf[0] = 0;
					length = 0;
					if(feof(authz)) break;
				}else{
					buf = realloc(buf, length + 2);
					buf[length] = ch;
					buf[length + 1] = 0;
					length++;
				}
			}
			free(buf);
			fclose(authz);
			authz = fopen(SWINEHUB_AUTHZ, "w");
			_FWRITE(authz, authzbuf);
			fclose(authz);
			free(authzbuf);
			remove(SWINEHUB_AUTHZ_LOCK);
			char* repofull = sstrcat3(result.dptr, "-", reponame);
			int pid = fork();
			if(pid == 0){
				chdir(config->path);
				execl("/bin/rm", "/bin/rm", "-rf", repofull, NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
			free(repofull);
			printf("Status: 200 OK\r\n");
			printf("\r\n");
			printf("{\"message\": \"Repository successfully deleted\", \"code\": %d}", SWINEHUB_CODE_OK);
		}else{
			printf("Status: 403 Forbidden\r\n");
			printf("\r\n");
			printf("{\"error\": \"Invalid token\", \"code\": %d}", SWINEHUB_CODE_INVALID_TOKEN);
		}

	}else{
		printf("Status: 400 Bad Request\r\n");
		printf("\r\n");
		printf("{\"error\": \"Unknown request\", \"code\": %d}", SWINEHUB_CODE_UNKNOWN_REQUEST);
	}
}
