/* $Id: repo.c 32 2023-04-26 16:49:22Z nishi $ */
#include <swinehub/repo.h>

#include <swinehub/base64.h>
#include <swinehub/config.h>
#include <swinehub/util.h>
#include <swinehub/user.h>
#include <swinehub/db.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>

#include <ndbm.h>

#define WRITE(fd, data) write(fd, data, strlen(data))

extern char* umethod;
extern char* swinehub_request;
extern char* swinehub_body;

EXTERN_DB;

void swinehub_repo(struct swinehub_config* config){
	char* slash = getenv("PATH_REPLACED") + 2;
	int i;
	char matched = 0;
	int index = -1;
	int index2 = -2;
	for(i = 0; slash[i] != 0; i++){
		if(slash[i] == '/'){
			if(index == -1) index = i;
			slash[i] = 0;
			matched = 1;
			if(index2 == -1){
				index2 = i;
				break;
			}
			index2 = -1;
		}
	}
	char* username = slash;
	char* reponame;
	char* svnpath = strdup("/");
	if(matched){
		username = slash;
		reponame = slash + index + 1;
		if(index2 != -1){
			free(svnpath);
			svnpath = sstrcat("/", slash + index2 + 1);
		}
	}
	if(strcmp(getenv("ISDAV"), "YES") == 0){
		printf("Status: 301 Moved Permanently\r\n");
		printf("Location: %s/%s-%s\r\n", config->prefix, username, reponame);
		printf("\r\n");
		return;
	}
	if(matched){
		if(strcmp(getenv("ISDAV"), "NO") == 0){
			int rev = -1;
			int status;
recheckout:
			{
				char* localpath = sstrcat5(config->path, "/", username, "-", reponame);
				struct stat st;
				if(stat(localpath, &st) != 0){
					free(localpath);
					printf("Repository not found.\n");
					return;
				}
				free(localpath);
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(config->cachepath);
					char* tmp = sstrcat4("file://", config->path, "/", username);
					char* checkoutpath = sstrcat3(tmp, "-", reponame);
				free(tmp);
					execl(config->svn, config->svn, "co", checkoutpath, NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
			}
			{
				char* tmp = sstrcat5(config->cachepath, "/", username, "-", reponame);
				char* chdirpath = strdup(tmp);
				free(tmp);
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(chdirpath);
					execl(config->svn, config->svn, "up", NULL);
					_exit(-1);
				}else{
					waitpid(pid, &status, 0);
				}
				free(chdirpath);
			}
			if(WIFEXITED(status)){
				if(WEXITSTATUS(status) != 0){
					char* tmp = sstrcat5(config->cachepath, "/", username, "-", reponame);
					char* rmpath = strdup(tmp);
					free(tmp);
					int pid = fork();
					if(pid == 0){
						int nullfd = open("/dev/null", O_RDWR);
						dup2(nullfd, fileno(stdout));
						dup2(nullfd, fileno(stderr));
						execl("/bin/rm", "/bin/rm", "-rf", rmpath, NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					free(rmpath);
					goto recheckout;
				}
			}
			char* tmp = sstrcat3(config->path, "/", username);
			char* tmp2 = sstrcat3(tmp, "-", reponame);
			char* str = sstrcat(tmp2, "/db/current");
			free(tmp);
			free(tmp2);
			struct stat s;
			if(stat(str, &s) == 0){
				FILE* f = fopen(str, "r");
				char* buf = malloc(s.st_size + 1);
				buf[s.st_size] = 0;
				fread(buf, 1, s.st_size, f);
				rev = atoi(buf);
				printf("<input type=\"hidden\" value=\"%s\" id=\"repouser\">\n", username);
				printf("<input type=\"hidden\" value=\"%s\" id=\"reponame\">\n", reponame);
				printf("<input type=\"hidden\" value=\"%s\" id=\"svnprefix\">\n", config->prefix);
				printf("<input type=\"hidden\" value=\"%s\" id=\"path\">\n", svnpath);
				printf("<h2>Repository <code>%s</code> by <a href=\"%s/~%s\"><code>%s</code></a> (Revision %d): <code>%s</code><button id=\"delete-button\" onclick=\"swinehub_delete()\" style=\"display: none;\">Delete</button></h2>\n", reponame, getenv("SCRIPT_NAME"), username, username, rev, svnpath);
				printf("<hr>\n");
				printf("<h3>README<span id=\"set-readme\" style=\"display: none;\"> | Set README: <input type=\"file\" id=\"readme-file\" onchange=\"swinehub_readme()\"></span></h3>\n");
				printf("<pre><code>");
				char* tmppath = sstrcat5(config->path, "/", username, "-", reponame);
				char* readme = sstrcat(tmppath, "/README.txt");
				free(tmppath);
				FILE* freadme = fopen(readme, "r");
				if(freadme != NULL){
					char ch;
					while(1){
						fread(&ch, 1, 1, freadme);
						if(feof(freadme)) break;
						if(ch == '<'){
							printf("&lt;");
						}else if(ch == '>'){
							printf("&gt;");
						}else{
							printf("%c", ch);
						}
					}
					fclose(freadme);
				}
				printf("</code></pre>");
				printf("<hr>\n");
				if(strcmp(svnpath, "/") == 0){
					printf("<h3>Repository Latest Commit</h3>\n");
					printf("<pre><code id=\"repo-latest-commit\">Fetching the latest commit ...</code></pre>\n");
					printf("<hr>\n");
				}else{
					printf("<h3>File Latest Commit</h3>\n");
					printf("<pre><code id=\"latest-commit\">Fetching the latest commit ...</code></pre>\n");
					printf("<hr>\n");
				}
				char* _path = sstrcat5(config->cachepath, "/", username, "-", reponame);
				char* path = sstrcat(_path, svnpath);
				free(_path);
				struct stat st;
				if(stat(path, &st) == 0){
					if(S_ISDIR(st.st_mode)){
						printf("<table border=\"0\" id=\"files\">\n");
						printf("	<tr>\n");
						printf("		<th>Name</th>\n");
						printf("		<th style=\"width: 100px;\">Type</th>\n");
						printf("		<th style=\"width: 250px;\">Size</th>\n");
						printf("	</tr>\n");
						struct dirent* d;
						DIR* dir = opendir(path);
						if(dir != NULL){
							while((d = readdir(dir)) != NULL){
								if(strcmp(d->d_name, ".") == 0 || (strcmp(svnpath, "/") == 0 ? (strcmp(d->d_name, "..") == 0) : 0) || strcmp(d->d_name, ".svn") == 0) continue;
								char* filepath = sstrcat3(path, "/", d->d_name);
								struct stat s;
								if(stat(filepath, &s) == 0){
									char ch = ' ';
									char* start = sstrcat3("<a href=\"./", d->d_name, "/\">");
									char* end = strdup("</a>");
									char* type = strdup("File");
									if(S_ISDIR(s.st_mode)){
										ch = '/';
										free(type);
										type = strdup("Directory");
									}
									printf("<tr>\n");
									printf("	<td>%s%s%c%s</td>\n", start, d->d_name, ch, end);
									printf("	<td>%s</td>\n", type);
									printf("	<td>%d</td>\n", s.st_size);
									printf("</tr>\n");
									free(start);
									free(end);
									free(type);
								}
								free(filepath);
							}
							closedir(dir);
							printf("</table>\n");
						}
					}else if(S_ISREG(st.st_mode)){
						int status;
						int pipes[2];
						pipe(pipes);
						int pid = fork();
						if(pid == 0){
							close(pipes[0]);
							int nullfd = open("/dev/null", O_RDWR);
							dup2(pipes[1], fileno(stdout));
							dup2(nullfd, fileno(stderr));
							char* chdirpath = sstrcat5(config->cachepath, "/", username, "-", reponame);
							chdir(chdirpath);
							char* getpath = sstrcat(".", svnpath);
							execl(config->svn, config->svn, "propget", "svn:mime-type", getpath, NULL);
							free(getpath);
							free(chdirpath);
							_exit(-1);
						}else{
							close(pipes[1]);
							waitpid(pid, &status, 0);
							if(WIFEXITED(status)){
								int exitstat = WEXITSTATUS(status);
								if(exitstat != 0){
									printf("<pre><code>");
									FILE* f = fopen(path, "r");
									char ch;
									while(1){
										fread(&ch, 1, 1, f);
										if(feof(f)) break;
										printf("%c", ch);
									}
									fclose(f);
									printf("</code></pre>\n");
								}else{
									char* mime = malloc(1);
									mime[0] = 0;
									char* buf = malloc(2);
									buf[1] = 0;
									while(1){
										if(read(pipes[0], buf, 1) == 0) break;
										if(buf[0] == '\n') continue;
										char* tmpbuf = mime;
										mime = sstrcat(tmpbuf, buf);
										free(tmpbuf);
									}
									free(buf);
									printf("File cannot be shown. (MIME: <code>%s</code>)\n", mime);
									free(mime);
								}
							}
						}
					}
				}else{
					printf("<tr>\n");
					printf("	<td>File/Directory not found</td>\n");
					printf("	<td></td>\n");
					printf("</tr>\n");
				}
				free(path);
				fclose(f);
				free(buf);
			}else{
				printf("Repository not found.");
			}
			free(str);
		}else{
			char* svn_path = sstrcat5(config->path, "/", username, "/", reponame);
			struct stat st;
			if(stat(svn_path, &st) == 0){
				char goahead = 0;
				if(
						(strcmp(umethod, "GET") == 0)
					||	(strcmp(umethod, "PROPFIND") == 0)
					||	(strcmp(umethod, "OPTIONS") == 0)
					||	(strcmp(umethod, "REPORT") == 0)
				){
					goahead = 1;
				}else{
					if(strcmp(getenv("ISAUTH"), "NO") == 0){
						goahead = 0;
					}else{
						goahead = 1;
					}
				}
			}else{
				printf("Status: 404 Not Found\r\n");
				printf("\r\n");
			}
		}
	}else{
		swinehub_user(config);
	}
	free(svnpath);
}

char swinehub_check_repo(struct swinehub_config* config){
	char result = 0;
	char* path = strdup(getenv("PATH_REPLACED")) + 2;
	int i;
	char count = 0;
	for(i = 0; path[i] != 0; i++){
		if(path[i] == '/'){
			path[i] = '-';
			count++;
		}
		if(count == 2){
			path[i] = 0;
			break;
		}
	}
	if(count != 2){
		datum key;
		key.dptr = path;
		key.dsize = strlen(path) + 1;
		datum value = dbm_fetch(swinehub_user_db, key);
		if(value.dptr != NULL) result = 1;
	}else{
		char* svn_path = sstrcat3(config->path, "/", path);
		struct stat s;
		if(stat(svn_path, &s) == 0) result = 1;
	}
	free(path);
	return result;
}
